/*  zweic -- a compiler for Zwei
 *
 *  Stephane Micheloud & LAMP
 *
 *  $Id: Scanner.java 2168 2005-11-14 15:24:45 +0100 (Mon, 14 Nov 2005) michelou $
 */

package zweic;

import java.io.InputStreamReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;

/**
 * This class implements the scanner of the compiler.
 */
class Scanner implements Tokens {

    /**
     * This character represents the end of the input.
     */
    private static final char EOF_CH = java.lang.Character.MAX_VALUE;

    /**
     * The current token class.
     */
    protected int token;

    /**
     * The pushed-back token.
     */
    protected class State {
        int start;
        String chars;
        int token;
        State(int start, String chars, int token) {
            this.start = start;
            this.chars = chars;
            this.token = token;
        }
    }
    protected LinkedList/*<State>*/ pushedBack = new LinkedList();

    /**
     * The position of the first character of the current token.
     */
    protected int start;

    /**
     * The string representation of the current token. This variable
     * is only set if the current token class has several possible
     * textual representations.
     */
    protected String chars;

    /** A buffer for constructing string representations of tokens.
     */
    private StringBuffer buf;

    /** The current character.
     */
    private char ch;

    /**
     * The line and the column of the current token.
     */
    private int line = 1;
    private int column = 0;

    /**
     * The input stream providing the input characters.
     */
    private final InputStreamReader in;

    /**
     * A table that associates a keyword with its corresponding
     * token class.
     */
    private final HashMap/*<String, Integer>*/ keywords = new HashMap();
    private int getToken(String key) {
        Object value = keywords.get(key);
        return (value != null) ? ((Integer) value).intValue() : IDENT;
    }
    private void enterKeyword(String key, int value) {
        Object prev = keywords.put(key, new Integer(value));
        assert prev == null : "keyword \"" + key + "\" entered twice";
    }
    {
        enterKeyword("class", CLASS);
        // ...  complter ...
    }

    Scanner(InputStreamReader in) {
        this.in = in;
        this.buf = new StringBuffer();
        nextCh();
        nextToken();
    }

    /**
     * This method reads the next token and stores the token class
     * in variable 'token'. If the token representation is not unique
     * it will also leave a textual representation in variable 'chars'.
     * Whitespaces and comments are skipped by this method.
     */
    public void nextToken() {
        if (0 == pushedBack.size()) {
            // skip comments and whitespace characters

            // ...  complter ...

            // initialize the position of the current token
            start = Position.encode(line, column);
            // read the current token
            token = readToken();
        }
        else {
            State oldState = (State) pushedBack.removeFirst();
            start = oldState.start;
            chars = oldState.chars;
            token = oldState.token;
        }
    }

    /**
     * Return the next token in the stream. 
     */
    public int peekAhead() {
        if (0 == pushedBack.size()) {
            int    oldStart = start;
            String oldChars = chars;
            int    oldToken = token;

            nextToken();
            int tok = token;
            pushedBack.add(new State(start, chars, token));

            start = oldStart;
            chars = oldChars;
            token = oldToken;
            return tok;
        }
        else {
            State oldState = (State) pushedBack.getFirst();
            return oldState.token;	
        }
    }

    /**
     * Read the next token, store its representation (if its not
     * unique) in variable 'chars' and return the token class.
     */
    private int readToken() {
        // ...  complter ...

    }

    /**
     * Return a textual representation of the current token.
     */
    public String representation() {
        String representation = tokenClass(token);
        if (token == NUMBER || token == IDENT || token == STRING)
            representation += "(\"" + chars + "\")";
        return representation;
    }

    /**
     * Return the textual representation of the given token class.
     */
    public static String tokenClass(int token) {
        switch (token) {
        case EOF       : return "<eof>";
        case BAD       : return "<bad>";
        case IDENT     : return "ident";
        case NUMBER    : return "number";
        case STRING    : return "string";
        case LPAREN    : return "(";
        case RPAREN    : return ")";
        // ...  complter ...
        default:
            throw new Error("unknown token class " + token);
        }
    }

    /**
     * Puts the next character into 'ch' and updates the current position.
     */
    private void nextCh() {
        switch (ch) {
            case EOF_CH:
                return;
            case '\n':
                column = 1;
                line++;
                break;
            default:
                column++;
        }
        try {
            ch = (char)in.read();
            oldch = ((oldch == '\r') && (ch == '\n')) ? (char)in.read() : ch;
            ch = (oldch == '\r') ? '\n' : oldch;
        } catch (IOException e) {
            Report.fail(Position.encode(line, column), e.getMessage());
        }
    }

    private char oldch;
}
